import os

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

import pandas as pd
import numpy as np
from scipy import stats
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import Sequential, layers

# import tensorflow.compat.v1 as tf
# tf.disable_v2_behavior()

from sklearn.feature_selection import chi2
from sklearn.ensemble import RandomForestClassifier
from sklearn.svm import SVC
from sklearn.linear_model import LinearRegression
from sklearn.feature_selection import RFE

visibleSize = 500
hiddenSize = 1000


class nAE(keras.Model):

    def __init__(self):
        super(nAE, self).__init__()

        # Encoder
        self.fc1 = layers.Dense(128)
        self.fc2 = layers.Dense(hiddenSize)  # get mean prediction
        self.fc3 = layers.Dense(hiddenSize)

        # Decoder
        self.fc4 = layers.Dense(128)
        self.fc5 = layers.Dense(500)

    def encoder(self, x):
        h = tf.nn.sigmoid(self.fc1(x))
        # get mean
        mu = self.fc2(h)
        # get variance
        log_var = self.fc3(h)

        return mu, log_var

    def decoder(self, z):
        out = tf.nn.sigmoid(self.fc4(z))
        out = self.fc5(out)

        return out

    def reparameterize(self, mu, log_var):
        eps = tf.random.normal(log_var.shape)

        std = tf.exp(log_var) ** 0.5

        z = mu + std * eps

        return z

    def call(self, inputs, training=None):
        mu, log_var = self.encoder(inputs)

        z = self.reparameterize(mu, log_var)

        x_hat = self.decoder(z)

        return x_hat, mu, log_var, z


class Discriminator(keras.Model):

    def __init__(self):
        super(Discriminator, self).__init__()

        self.fc1 = layers.Dense(visibleSize)

        self.fc2 = layers.Dense(512)

        self.fc3 = layers.Dense(128)

        self.fc4 = layers.Dense(1)

    def call(self, inputs, training=None):
        x = self.fc1(inputs)

        x = tf.nn.sigmoid(x)

        x = self.fc2(x)

        x = tf.nn.sigmoid(x)

        x = self.fc3(x)

        x = tf.nn.sigmoid(x)

        x = self.fc4(x)

        return x


def KL(rho, rho_j):
    return rho * tf.log(rho / rho_j) + (1 - rho) * tf.log((1 - rho) / (1 - rho_j))


def AE(data_train=[], learning_rate=1, epochs=1, batch_size=1, feature_selection_type=""):
    data_train = data_train.astype(np.float32)
    train_db = tf.data.Dataset.from_tensor_slices(data_train)
    train_db = train_db.shuffle(batch_size * 5).batch(batch_size)

    model = nAE()
    model.build(input_shape=(4, visibleSize))

    discriminator = Discriminator()
    discriminator.build(input_shape=(None, visibleSize))

    model.summary()
    optimizer = tf.optimizers.Adam(learning_rate)

    discriminator.summary()
    dis_optimizer = tf.optimizers.Adam(learning_rate)
    for epoch in range(epochs):

        for step, x in enumerate(train_db):

            x = tf.reshape(x, [-1, 500])
            with tf.GradientTape() as tape:
                x_rec_logits, mu, log_var, h = model(x)

                rec_loss = tf.nn.sigmoid_cross_entropy_with_logits(labels=x, logits=x_rec_logits)
                rec_loss = tf.reduce_sum(rec_loss) / x.shape[0]

                kl_div = -0.5 * (log_var + 1 - mu ** 2 - tf.exp(log_var))
                kl_div = tf.reduce_sum(kl_div) / x.shape[0]

                loss = rec_loss + 1. * kl_div

            grads = tape.gradient(loss, model.trainable_variables)
            optimizer.apply_gradients(zip(grads, model.trainable_variables))

            with tf.GradientTape() as tape:

                fake_info = tf.random.uniform([x.shape[0], x.shape[1]], minval=-1., maxval=1)
                fake_info, _, _, _ = model(fake_info)
                fake_logits = discriminator(fake_info)
                real_info, _, _, _ = model(x)
                real_logits = discriminator(real_info)

                fake_loss = tf.nn.sigmoid_cross_entropy_with_logits(logits=fake_logits,
                                                                    labels=tf.zeros_like(fake_logits))
                real_loss = tf.nn.sigmoid_cross_entropy_with_logits(logits=real_logits,
                                                                    labels=tf.ones_like(real_logits))
                fake_loss = tf.reduce_mean(fake_loss)
                real_loss = tf.reduce_mean(real_loss)

                d_loss = fake_loss + real_loss

            grads = tape.gradient(d_loss, discriminator.trainable_variables)
            dis_optimizer.apply_gradients(zip(grads, discriminator.trainable_variables))

            if step % 100 == 0:
                print(epoch, step, 'kl div:', float(kl_div), 'rec loss:', float(rec_loss), " dloss:", float(d_loss))

    model.save_weights('./result/model/model')
    discriminator.save_weights('./result/model/model0')

    hid = None
    for step, x in enumerate(train_db):
        if (step == 0):
            x_rec_logits, mu, log_var, hid = model(x)
            continue
        x_rec_logits, mu, log_var, h = model(x)
        hid = tf.concat([hid, h], axis=0)

    return hid


def Ttest(X, Y, ttest_num):
    Y1 = np.where(Y == 1)[0]
    Y0 = np.where(Y == 0)[0]
    p_list = []
    for i in range(len(X[0])):
        p_l = stats.levene(X[Y1, i], X[Y0, i])[1]
        equal_var = [True if p_l > 0.05 else False]
        p_list.append(stats.ttest_ind(X[Y1, i], X[Y0, i], equal_var=equal_var)[1])
    T_index = np.argsort(np.array(p_list))[0:ttest_num]
    X1 = X[:, T_index]
    return T_index


def Wtest(X, Y, wtest_num):
    Y1 = np.where(Y == 1)[0]
    Y0 = np.where(Y == 0)[0]
    p_list = []
    for i in range(len(X[0])):
        p_list.append(stats.ranksums(X[Y1, i], X[Y0, i])[1])
    W_index = np.argsort(np.array(p_list))[0:wtest_num]
    return W_index


def Chi2(X, Y, ctest_num):
    p_list = chi2(X, Y)[1]
    C_index = np.argsort(np.array(p_list))[0:ctest_num]
    return C_index


def RF(X, Y, rtest_num):
    forest = RandomForestClassifier(random_state=0, n_jobs=1)
    forest.fit(X, Y)
    importance = forest.feature_importances_
    R_index = np.argsort(np.array(importance))[::-1][0:rtest_num]
    return R_index


def LR_RFE(X, Y, lrtest_num):
    clf = LinearRegression()
    rfe = RFE(clf, n_features_to_select=1)
    rfe.fit(X, Y)
    rank = rfe.ranking_
    LR_index = np.argsort(np.array(rank))[0:lrtest_num]
    return LR_index


def SVM_RFE(X, Y, srtest_num):
    clf = SVC(kernel='linear', random_state=0)
    rfe = RFE(clf, n_features_to_select=1)
    rfe.fit(X, Y)
    rank = rfe.ranking_
    SR_index = np.argsort(np.array(rank))[0:srtest_num]
    return SR_index


def train(feature_selection_type, train_data, train_label):
    X = np.array(train_data)

    train_sae = AE(data_train=X, learning_rate=0.0002, epochs=400, batch_size=128,
                   feature_selection_type=feature_selection_type)

    pd.DataFrame(train_sae).to_csv("./result/" + feature_selection_type + "_train_sae.csv")


if __name__ == '__main__':
    feature_selection_type = "Ttest"

    train_data = pd.read_csv("./result/sample/data_train.csv", index_col=0)
    train_label = pd.read_csv("./result/sample/label_train.csv", index_col=0)
    train(feature_selection_type, train_data, train_label)
